#include <GlewOpenGL.h>
#include "ProgGLSL.h"
#include "Uniform.h"
#include "UniformGLSL.h"
#include "ProgParam.h"
#include "ProgGLSL.h"
#include "Message.h"
#include "Util.h"
#include <QInputDialog>
#include <QDir>

namespace apig {

/**************************************************************************************/

//ProgGLSL::ProgGLSL() : linked(false) {

ProgGLSL* ProgGLSL::create() {
    GLuint id = glCreateProgram();
    if (id == 0) {
        Message::error("ProgGLSL : impossible de creer le programme");
        return NULL;
        }
    return new ProgGLSL(id);
    }

ProgGLSL::~ProgGLSL() {
    foreach (ShaderGLSL s, shaders) { detach(s); s.destroy(); }
    //shaders.clear();
    glDeleteProgram(id);
    clearParameters();
    }

void ProgGLSL::attach(ShaderGLSL shader) { glAttachShader(id, shader.id); shaders.append(shader); }
void ProgGLSL::detach(ShaderGLSL shader) { glDetachShader(id, shader.id); shaders.removeAll(shader); }

void ProgGLSL::link() {
    glLinkProgram(id);
    
    if (getProgParam(GL_LINK_STATUS) == GL_FALSE)
        Message::error(QString("- erreur : l'edition de liens a echoue :\n%1").arg(getInfoLog()), false);
    else
        linked = true;
    }

void ProgGLSL::validate() {
    glValidateProgram(id);
    
    if (getProgParam(GL_VALIDATE_STATUS) == GL_FALSE)
        Message::info(QString("> programme non valide :\n%1").arg(getInfoLog()));
    else
        Message::info("> programme valide.");
    }

void ProgGLSL::activate() {
    glUseProgram(id);
    updateVariables();
    textures.activate();
    }

void ProgGLSL::deactivate() {
    glUseProgram(0);
    textures.deactivate();
    }

/**************************************************************************************/

void ProgGLSL::clearParameters() {
    textures.clear();
    while (!parameters.isEmpty()) delete parameters.takeFirst();
    }

void ProgGLSL::updateVariables()  const {
    foreach (ProgParam *param, parameters) param->update();
    }

void ProgGLSL::activateTextures() const {
    textures.activate();
    }

bool ProgGLSL::isLinked() const { return linked; }

/**************************************************************************************/

UniformGLSL* ProgGLSL::createUniformGLSL(QString name, bool warn) const {
    UniformGLSL* param = new UniformGLSL(name, id);
    if (!param->exists()) {
        if (warn) Message::error(QString("<%1> n'est pas une variable uniform active du programme").arg(name));
        delete param;
        param = NULL;
        }
    return param;
    }

void ProgGLSL::addUniform(Uniform *uniform, bool warn, bool ownUniform) {
    UniformGLSL *param = createUniformGLSL(uniform->name, warn);
    if (param != NULL) {
        ProgParam *p = new ProgParam(uniform, param, ownUniform);
        parameters.append(p);
        }
    }

void ProgGLSL::addTexture(QString name, Texture *texture, bool warn) {
    int texNum = textures.add(texture);
    addUniform(new Sampler(name, texture->getID(), texNum), warn, true);
    }

void ProgGLSL::addUniform(QString name, const float *value, bool warn)     { addUniform(new UniformFloat(name,  value), warn, true); }
void ProgGLSL::addUniform(QString name, const bool *value, bool warn)      { addUniform(new UniformBool(name,   value), warn, true); }
void ProgGLSL::addUniform(QString name, const int *value, bool warn)       { addUniform(new UniformInt(name,    value), warn, true); }
void ProgGLSL::addUniformVec2(QString name, const float *value, bool warn) { addUniform(new UniformVec2(name,   value), warn, true); }
void ProgGLSL::addUniformVec3(QString name, const float *value, bool warn) { addUniform(new UniformVec3(name,   value), warn, true); }
void ProgGLSL::addUniformVec4(QString name, const float *value, bool warn) { addUniform(new UniformVec4(name,   value), warn, true); }
void ProgGLSL::addUniformIVec2(QString name, const int *value, bool warn)  { addUniform(new UniformIVec2(name,  value), warn, true); }
void ProgGLSL::addUniformIVec3(QString name, const int *value, bool warn)  { addUniform(new UniformIVec3(name,  value), warn, true); }
void ProgGLSL::addUniformIVec4(QString name, const int *value, bool warn)  { addUniform(new UniformIVec4(name,  value), warn, true); }
void ProgGLSL::addUniformBVec2(QString name, const bool *value, bool warn) { addUniform(new UniformBVec2(name,  value), warn, true); }
void ProgGLSL::addUniformBVec3(QString name, const bool *value, bool warn) { addUniform(new UniformBVec3(name,  value), warn, true); }
void ProgGLSL::addUniformBVec4(QString name, const bool *value, bool warn) { addUniform(new UniformBVec4(name,  value), warn, true); }
void ProgGLSL::addUniformMat2(QString name, const float *value, bool warn) { addUniform(new UniformMat2(name,   value), warn, true); }
void ProgGLSL::addUniformMat3(QString name, const float *value, bool warn) { addUniform(new UniformMat3(name,   value), warn, true); }
void ProgGLSL::addUniformMat4(QString name, const float *value, bool warn) { addUniform(new UniformMat4(name,   value), warn, true); }

/**************************************************************************************/

#define CREATE_PARAM_AND_CALL(functionCall) {           \
    UniformGLSL *param = createUniformGLSL(name, warn); \
    if (param != NULL) {                                \
        param->functionCall;                            \
        delete param;                                   \
        }                                               \
    }
    
void ProgGLSL::setUniform(QString name, float value, bool warn)                                  CREATE_PARAM_AND_CALL(setUniform1f(value))
void ProgGLSL::setUniform(QString name, int value, bool warn)                                    CREATE_PARAM_AND_CALL(setUniform1i(value))
void ProgGLSL::setUniform(QString name, bool value, bool warn)                                   CREATE_PARAM_AND_CALL(setUniform1b(value))

void ProgGLSL::setUniformVec2(QString name, float v0, float v1, bool warn)                       CREATE_PARAM_AND_CALL(setUniform2f(v0, v1))
void ProgGLSL::setUniformVec3(QString name, float v0, float v1, float v2, bool warn)             CREATE_PARAM_AND_CALL(setUniform3f(v0, v1, v2))
void ProgGLSL::setUniformVec4(QString name, float v0, float v1, float v2, float v3, bool warn)   CREATE_PARAM_AND_CALL(setUniform4f(v0, v1, v2, v3))
void ProgGLSL::setUniformVec2(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform2fv(value))
void ProgGLSL::setUniformVec3(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform3fv(value))
void ProgGLSL::setUniformVec4(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform4fv(value))

void ProgGLSL::setUniformIVec2(QString name, int v0, int v1, bool warn)                          CREATE_PARAM_AND_CALL(setUniform2i(v0, v1))
void ProgGLSL::setUniformIVec3(QString name, int v0, int v1, int v2, bool warn)                  CREATE_PARAM_AND_CALL(setUniform3i(v0, v1, v2))
void ProgGLSL::setUniformIVec4(QString name, int v0, int v1, int v2, int v3, bool warn)          CREATE_PARAM_AND_CALL(setUniform4i(v0, v1, v2, v3))
void ProgGLSL::setUniformIVec2(QString name, const int *value, bool warn)                        CREATE_PARAM_AND_CALL(setUniform2iv(value))
void ProgGLSL::setUniformIVec3(QString name, const int *value, bool warn)                        CREATE_PARAM_AND_CALL(setUniform3iv(value))
void ProgGLSL::setUniformIVec4(QString name, const int *value, bool warn)                        CREATE_PARAM_AND_CALL(setUniform4iv(value))

void ProgGLSL::setUniformBVec2(QString name, bool v0, bool v1, bool warn)                        CREATE_PARAM_AND_CALL(setUniform2b(v0, v1))
void ProgGLSL::setUniformBVec3(QString name, bool v0, bool v1, bool v2, bool warn)               CREATE_PARAM_AND_CALL(setUniform3b(v0, v1, v2))
void ProgGLSL::setUniformBVec4(QString name, bool v0, bool v1, bool v2, bool v3, bool warn)      CREATE_PARAM_AND_CALL(setUniform4b(v0, v1, v2, v3))
void ProgGLSL::setUniformBVec2(QString name, const bool *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform2bv(value))
void ProgGLSL::setUniformBVec3(QString name, const bool *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform3bv(value))
void ProgGLSL::setUniformBVec4(QString name, const bool *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniform4bv(value))

void ProgGLSL::setUniformMat2(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniformMatrix2fv(value))
void ProgGLSL::setUniformMat3(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniformMatrix3fv(value))
void ProgGLSL::setUniformMat4(QString name, const float *value, bool warn)                       CREATE_PARAM_AND_CALL(setUniformMatrix4fv(value))

/**************************************************************************************/

namespace {
    /*
    QString filePath(QString dirName, QString fileName) {
        return QDir::cleanPath(QDir(dirName).filePath(fileName));
        }
    */
    
    QString extension(ShaderGLSL::Type type) {
        return type == ShaderGLSL::VERTEX ? ".vert" : ".frag";
        }
    /*
    QStringList subDirsList(QString dirName) {
        QDir dir(dirName);
        return dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot, QDir::Name);
        }
    */
    QString inputSubDir(QString dirName) {
        QStringList list = util::subDirsList(dirName);
        bool ok;
        const QString subDirName = QInputDialog::getItem(NULL, "Repertoires de shaders", "Choix", list, 0, false, &ok);
        return ok ? subDirName : QString::null;
        }
    
    QString inputSourceFile(QString dirName, ShaderGLSL::Type type) {       // retourne un nom de fichier relatif
        QStringList list = ProgGLSL::sourcesList(dirName, type, true);
        bool ok;
        const QString subDirName = QInputDialog::getItem(NULL, "Repertoires de shaders", "Choix", list, 0, false, &ok);
        return ok ? subDirName : QString::null;
        }
    }

/**************************************************************************************/

QStringList ProgGLSL::sourcesList(QString dirName, ShaderGLSL::Type type, bool relative) {
    QString filter = "*" + extension(type);
    QDir dir(dirName, filter);
    QStringList list = dir.entryList(QDir::Files, QDir::Name);
    
    if (!relative)
        for (int i=0; i<list.size(); i++)
            list[i] = util::filePath(dirName, list[i]);
    
    return list;
    }

/**************************************************************************************/

void ProgGLSL::bindShaderDir(QString dirName, ShaderGLSL::Type shaderType) {
    QStringList sources = sourcesList(dirName, shaderType, false);
    //QStringList sources = locator.sources(dirName, shaderType, true, false, true);
    if (sources.isEmpty())
        Message::error(QString("pas de %1 shader dans le repertoire '%2'").arg(shaderType == ShaderGLSL::VERTEX ? "vertex" : "fragment").arg(dirName));
    else foreach (QString s, sources)
        bindShaderFile(s, shaderType);
    }

void ProgGLSL::bindShaderFile(QString fileName, ShaderGLSL::Type type) {
    ShaderGLSL shader(fileName, type, true);
    shader.compile();
    attach(shader);
    }

void ProgGLSL::bindShaderSource(QString source, ShaderGLSL::Type type) {
    ShaderGLSL shader(source, type, false);
    shader.compile();
    attach(shader);
    }

/**************************************************************************************/
// methodes de chargement de sources :

ProgGLSL* ProgGLSL::loadSubDir(QString dirName) {
    QString subDirName = inputSubDir(dirName);
    if (subDirName.isNull()) return NULL;
    return loadDir(util::filePath(dirName, subDirName));
    }

ProgGLSL* ProgGLSL::loadDir(QString dirName) {
    ProgGLSL* prog = create();
    if (prog != NULL) {
        prog->bindShaderDir(dirName, ShaderGLSL::VERTEX);      // chargement des vertex shaders
        prog->bindShaderDir(dirName, ShaderGLSL::FRAGMENT);    // chargement des fragment shaders
        prog->link();
        }
    return prog;
    }

ProgGLSL* ProgGLSL::load(QString name) {
    ProgGLSL* prog = create();
    if (prog != NULL) {
        prog->bindShaderFile(name + extension(ShaderGLSL::VERTEX),   ShaderGLSL::VERTEX);
        prog->bindShaderFile(name + extension(ShaderGLSL::FRAGMENT), ShaderGLSL::FRAGMENT);
        prog->link();
        }
    return prog;
    }

ProgGLSL* ProgGLSL::load(QString vertName, QString fragName) {
    ProgGLSL* prog = create();
    if (prog != NULL) {
        prog->bindShaderFile(vertName, ShaderGLSL::VERTEX);
        prog->bindShaderFile(fragName, ShaderGLSL::FRAGMENT);
        prog->link();
        }
    return prog;
    }

ProgGLSL* ProgGLSL::loadFrag(QString dirName, QString fragName) {
    QStringList vertList = sourcesList(dirName, ShaderGLSL::VERTEX, false);
    if (vertList.size() != 1) {
        Message::error(QString("le repertoire '%1' doit contenir un unique vertex shader").arg(dirName));
        return NULL;
        }
        
    // pas de nom specifie, l'utilisateur doit en choisir un :
    if (fragName.isEmpty()) fragName = inputSourceFile(dirName, ShaderGLSL::FRAGMENT);
    if (fragName.isEmpty()) return NULL;
        
    return load(vertList.first(), util::filePath(dirName, fragName));
    }

/**************************************************************************************/

GLint ProgGLSL::getProgParam(GLenum paramName) const { int res; glGetProgramiv(id, paramName, &res); return res; }

QString ProgGLSL::getInfoLog() const {
    int logLength = getProgParam(GL_INFO_LOG_LENGTH);
    char *infoLog = new char[logLength];
    glGetProgramInfoLog(id, logLength, NULL, infoLog);
    QString res(infoLog);
    delete infoLog;
    return res;
    }

/**************************************************************************************/

void ProgGLSL::printInfo() const {
    Message::info("> attributs actifs :");
    printActiveAttribs();
    Message::info("> uniforms actifs :");
    printActiveUniforms();
    }
    
void ProgGLSL::printActiveAttribs() const {
    int numUniforms = getProgParam(GL_ACTIVE_ATTRIBUTES);
    int maxLength   = getProgParam(GL_ACTIVE_ATTRIBUTE_MAX_LENGTH);
    char *name = new char[maxLength];
    for (int i=0; i<numUniforms; i++) {
        int length, size;
        GLenum type;
        glGetActiveAttrib(id, i, maxLength, &length, &size, &type, name);
        QString t = typeEnumString(type);
        if (size > 1) t += QString("[%1]").arg(size);
        Message::info(QString("  - %1 %2").arg(t).arg(name));
        }
    delete[] name;
    }

void ProgGLSL::printActiveUniforms() const {
    int numUniforms = getProgParam(GL_ACTIVE_UNIFORMS);
    int maxLength   = getProgParam(GL_ACTIVE_UNIFORM_MAX_LENGTH);
    char *name = new char[maxLength];
    for (int i=0; i<numUniforms; i++) {
        int length, size;
        GLenum type;
        glGetActiveUniform(id, i, maxLength, &length, &size, &type, name);
        QString t = typeEnumString(type);
        if (size > 1) t += QString("[%1]").arg(size);
        Message::info(QString("  - %1 %2").arg(t).arg(name));
        }
    delete[] name;
    }

QString ProgGLSL::typeEnumString(GLenum type) {
    #define _(e) if (type == e) return #e;
    switch (type) {
        case GL_FLOAT       : return "float";
        case GL_FLOAT_VEC2  : return "vec2";
        case GL_FLOAT_VEC3  : return "vec3";
        case GL_FLOAT_VEC4  : return "vec4";
        case GL_INT         : return "int";
        case GL_INT_VEC2    : return "ivec2";
        case GL_INT_VEC3    : return "ivec3";
        case GL_INT_VEC4    : return "ivec4";
        case GL_BOOL        : return "bool";
        case GL_BOOL_VEC2   : return "bvec2";
        case GL_BOOL_VEC3   : return "bvec3";
        case GL_BOOL_VEC4   : return "bvec4";
        case GL_FLOAT_MAT2  : return "mat2";
        case GL_FLOAT_MAT3  : return "mat3";
        case GL_FLOAT_MAT4  : return "mat4";
        case GL_SAMPLER_1D  : return "sampler1D";
        case GL_SAMPLER_2D  : return "sampler2D";
        case GL_SAMPLER_3D  : return "sampler3D";          
        case GL_SAMPLER_CUBE        : return "samplerCube";
        case GL_SAMPLER_1D_SHADOW   : return "sampler1DShadow";
        case GL_SAMPLER_2D_SHADOW   : return "sampler2DShadow";
        default : return "<type inconnu>";
        }
    }

}

