#ifndef __VEC4__
#define __VEC4__

#include "Vec3.h"
#include <QString>
#include <QDomElement>
#include <QDomDocument>

namespace apig {

    /** Representation of a four-dimensional vector.
     * It can be used for homogeneous coordinates manipulation.
     */
    class Vec4 {
        public:
            Vec4(float x, float y, float z, float w = 1);
            Vec4(); // vecteur (0,0,0,1)
            Vec4(float c); // vecteur (c,c,c,c)
            Vec4(Vec3 v, float w = 1);
            Vec4(const float *v);     // tableau de 4 valeurs
            //Vec4(qglviewer::Vec vec);
            Vec4(const QDomElement &element);
            
        // conversions :
        //--------------
            // operateurs de conversion automatique :
            inline float& operator[](int i) { return v[i]; }
            inline operator const float* () const { return v; }       // cast du Vec4 en const float*
            // autres conversions :
            Vec3    toVec3()    const;
            QString toQString() const;
            // lecture / ecriture d'un element XML :
            QDomElement domElement(const QString &name, QDomDocument &document) const;
            void initFromDOMElement(const QDomElement &element);
    
        // operations geometriques :
        //--------------------------
            friend Vec4 dir(const Vec4 &a, const Vec4 &b);    // vecteur representant la direction de <a> a <b>
            void homogenize();      // division de toutes les composante par la derniere
            // coordonnees normalisees :
            float xn() const { return x / w; }
            float yn() const { return y / w; }
            float zn() const { return z / w; }
            
        // operateurs arithmetiques :
        //---------------------------
            // operateurs arithmetiques unaires :
            friend Vec4 operator-(const Vec4 &a);
    
            // operateurs arithmetiques binaires :
            friend Vec4 operator+(const Vec4 &a, const Vec4 &b);
            friend Vec4 operator-(const Vec4 &a, const Vec4 &b);
            friend Vec4 operator*(const Vec4 &a, const Vec4 &b);
            friend Vec4 operator/(const Vec4 &a, const Vec4 &b);
                
            // operations avec un scalaire :
            friend Vec4 operator*(float s, const Vec4 &a);
            friend Vec4 operator*(const Vec4 &a, float s);
            friend Vec4 operator/(float s, const Vec4 &a);
            friend Vec4 operator/(const Vec4 &a, float s);
            
            // operateurs de modification :
            Vec4& operator+=(const Vec4 &a);
            Vec4& operator-=(const Vec4 &a);
            Vec4& operator*=(const Vec4 &a);
            Vec4& operator/=(const Vec4 &a);
            Vec4& operator*=(float s);
            Vec4& operator/=(float s);
        
        // operateurs de comparaison :
        //----------------------------
            friend bool operator==(const Vec4 &a, const Vec4 &b);   // comparaison a 1e-10 pres
            friend bool operator!=(const Vec4 &a, const Vec4 &b);   // comparaison a 1e-10 pres
            // les comparaisons suivantes sont le <et> booleen des comparaisons sur les composantes :
            friend bool operator>=(const Vec4 &a, const Vec4 &b);
            friend bool operator<=(const Vec4 &a, const Vec4 &b);
            friend bool operator>(const Vec4 &a, const Vec4 &b);
            friend bool operator<(const Vec4 &a, const Vec4 &b);
            friend bool operator>=(const Vec4 &a, float b);
            friend bool operator<=(const Vec4 &a, float b);
            friend bool operator>(const Vec4 &a, float b);
            friend bool operator<(const Vec4 &a, float b);
        
        // autres fonctions pratiques :
        //-----------------------------
            friend Vec4 abs(const Vec4 &a);
            friend Vec4 sign(const Vec4 &a);
            friend Vec4 floor(const Vec4 &a);
            friend Vec4 ceil(const Vec4 &a);
            friend Vec4 fract(const Vec4 &a);
            friend Vec4 min(const Vec4 &a, const Vec4 &b);
            friend Vec4 min(const Vec4 &a, float b);
            friend Vec4 max(const Vec4 &a, const Vec4 &b);
            friend Vec4 max(const Vec4 &a, float b);
            friend Vec4 clamp(const Vec4 &a, const Vec4 &min, const Vec4 &max);
            friend Vec4 clamp(const Vec4 &a, float min, float max);
            friend Vec4 mix(const Vec4 &a, const Vec4 &b, const Vec4 &alpha);     // (1-alpha) * a + alpha * b
            friend Vec4 mix(const Vec4 &a, const Vec4 &b, float alpha);           // (1-alpha) * a + alpha * b
            friend Vec4 step(const Vec4 &e, const Vec4 &a);
            friend Vec4 step(float e, const Vec4 &a);
            friend float min(const Vec4 &a);    // minimum des composantes
            friend float max(const Vec4 &a);    // maximum des composantes
        
        // operateurs de swizzle :
        //------------------------
            Vec3 xyz() const { return Vec3(x,y,z); }
            
        // calcul de vecteurs aleatoires (variables uniformes) :
        //------------------------------------------------------
            static Vec4 random();                       // valeurs dans [0,1[
            static Vec4 random(float min, float max);   // valeurs dans [min,max[
            static Vec4 random(Vec4 min, Vec4 max);     // bornes par composantes
    
        // operations OpenGL :
        //--------------------
            void glVertex() const;    // effectue le glVertex correspondant
            
        private:
            union {
                struct { float x, y, z, w; };
                float v[4];
                };
        };
    
    }

#endif

